// AviDialog.cpp : implementation file
//

#include "headers.h"
#include "resource.h"
#include "AviDialog.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// AviDialog dialog



AviDialog::AviDialog(CWnd* pParent /*=NULL*/)
	: CDialog(AviDialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(AviDialog)
	filename = _T("");
	pathname = _T("");
	avidims = _T("");
	aviframes = _T("");
	//}}AFX_DATA_INIT

	index=-1;
}


void AviDialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(AviDialog)
	DDX_Control(pDX, IDC_MODELIST, modelist);
	DDX_Text(pDX, IDC_FILENAME, filename);
	DDX_Text(pDX, IDC_PATH, pathname);
	DDX_Text(pDX, IDC_AVISDIMS, avidims);
	DDX_Text(pDX, IDC_AVIFRAMES, aviframes);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(AviDialog, CDialog)
	//{{AFX_MSG_MAP(AviDialog)
	ON_BN_CLICKED(IDC_OPENAVI, OnOpenAvi)
	ON_LBN_SELCHANGE(IDC_MODELIST, OnSelchangeModelist)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

void AviDialog::SetArray( DisplayModeArray* d )
{
	displaymodelist=d;
}

/////////////////////////////////////////////////////////////////////////////
// AviDialog message handlers

void AviDialog::OnOpenAvi() 
{
	SetCurrentDirectory("c:\\avi");

	static char BASED_CODE filter[] = "AVI Files (*.avi)|*.avi||";
	CFileDialog opendialog( TRUE, 0, 0, OFN_FILEMUSTEXIST, filter, this );
	if ( opendialog.DoModal() == IDOK )
	{
		filename = opendialog.GetFileName();
		fullfilename=opendialog.GetPathName();
		int pathlen=fullfilename.ReverseFind( '\\' );
		pathname=fullfilename.Left( pathlen );

		InitControls();
		UpdateData(FALSE);
	}

}

void AviDialog::OnSelchangeModelist() 
{
	int cursel=modelist.GetCurSel();
	ASSERT(cursel!=LB_ERR);

	index=cursel;

	if (fullfilename!="")
		GetDlgItem(IDOK)->EnableWindow();
}

void AviDialog::OnOK() 
{
	index=modelist.GetCurSel();
	if (index==LB_ERR)
		index=-1;
	
	CDialog::OnOK();
}

BOOL AviDialog::GetAviSpecs( LPCTSTR filename, int* w, int* h, int* d, int* f )
{
	long r;

	PAVISTREAM avistream;

    r=AVIStreamOpenFromFile( &avistream, filename,
		streamtypeVIDEO, 0, OF_READ | OF_SHARE_EXCLUSIVE, 0 );
	TRACE("AVIStreamOpenFromFile: %s\n", r==0 ? "OK" : "failed" );

	long fmtlen;
    r=AVIStreamFormatSize( avistream, 0, &fmtlen );
	TRACE("AVIStreamFormatSize: %s\n", r==0 ? "OK" : "failed" );

	LPBITMAPINFOHEADER format;
	format = (LPBITMAPINFOHEADER)new BYTE[fmtlen];
	ZeroMemory( format, fmtlen );

    r=AVIStreamReadFormat( avistream, 0, format, &fmtlen );
	TRACE("AVIStreamReadFormat: %s\n", r==0 ? "OK" : "failed" );

	*w=format->biWidth;
	*h=format->biHeight;
	*d=format->biBitCount;
	TRACE("Dialog: w=%d h=%d d=%d\n", *w, *h, *d);

	*f=AVIStreamLength( avistream );

	AVIStreamRelease( avistream );

	delete [] format;

	return TRUE;
}

BOOL AviDialog::OnInitDialog() 
{
	TRACE("OnInitDialog()...\n");
	CDialog::OnInitDialog();

	modelist.ResetContent();

	int size=displaymodelist->GetSize();
	for (int i=0;i<size;i++)
		modelist.InsertString( i, displaymodelist->GetAt(i).desc );

	GetDlgItem(IDOK)->EnableWindow( FALSE );
	
	if (fullfilename!="")
		InitControls();
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

BOOL AviDialog::InitControls()
{
	GetDlgItem(IDOK)->EnableWindow( FALSE );

	int w, h, depth, frames;
	if (GetAviSpecs( fullfilename, &w, &h, &depth, &frames )==FALSE)
		return TRUE;

	avidims.Format( "%dx%d", w, h );
	aviframes.Format( "%d", frames );

	if (index!=-1 && modelist.GetCount()>0)
	{
		modelist.SetCurSel( index );
		GetDlgItem(IDOK)->EnableWindow( TRUE );
	}

	return TRUE;
}

