#include "Headers.h"
#include "resource.h"
#include "AviDialog.h"
#include "AviPlayWin.h"

#pragma comment (lib,"ddraw.lib")
#pragma comment (lib,"dxguid.lib")
#pragma comment (lib,"vfw32.lib")


BEGIN_MESSAGE_MAP(AviPlayWin, DirectDrawWin)
	//{{AFX_MSG_MAP(AviPlayWin)
	ON_WM_KEYDOWN()
	ON_WM_RBUTTONDOWN()
	ON_WM_DESTROY()
	ON_WM_CREATE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


AviPlayWin::AviPlayWin()
{
	syspal=0;
	avipal=0;

	decomp=0;

	avisurf=0;
	avidialog=0;
	x=0;
	y=0;

	avistream=0;
	srcfmt=0;
	dstfmt=0;

	rawdata=0;
	finaldata=0;

	startframe=0;
	endframe=0;
	curframe=0;
}

void AviPlayWin::DrawScene()
{
	long r;
	r=AVIStreamRead( avistream, curframe, 1, rawdata, buflen, 0/*&bytes*/, 0 );
	if (r)
	{
		TRACE("AVIStreamRead failed: ");
		switch (r)
		{
			case AVIERR_BUFFERTOOSMALL:
				TRACE("BUFFERTOOSMALL\n");
				break;
			case AVIERR_MEMORY:
				TRACE("MEMORY\n");
				break;
			case AVIERR_FILEREAD:
				TRACE("FILEREAD\n");
				break;
		}
	}

	r=ICDecompress( decomp, 0, srcfmt, rawdata, dstfmt, finaldata);
	if (r != ICERR_OK)
		TRACE("ICDecompress: failed (error code %d)\n", r);
	
	curframe=(curframe<endframe) ? curframe+1 : startframe;

	UpdateAviSurface();

	backsurf->BltFast( x, y, avisurf, 0, DDBLTFAST_WAIT );

	primsurf->Flip( 0, DDFLIP_WAIT );
}

BOOL AviPlayWin::CreateAviSurface()
{
	if (avisurf)
		avisurf->Release(), avisurf=0;

	avisurf=CreateSurface( srcfmt->biWidth, srcfmt->biHeight );

	CRect displayrect=GetDisplayRect();

	x=(displayrect.Width()-srcfmt->biWidth)/2;
	y=0;

	return TRUE;
}

BOOL AviPlayWin::UpdateAviSurface()
{
	HRESULT r;

	if (finaldata==0)
		return FALSE;

	DWORD dwWidth = (srcfmt->biWidth+3) & ~3;
	DWORD dwHeight = srcfmt->biHeight;

    DDSURFACEDESC desc;
	ZeroMemory( &desc, sizeof(desc) );
	desc.dwSize = sizeof(desc);

	r = avisurf->Lock( 0, &desc, DDLOCK_WAIT, 0 );
	if (r==DD_OK)
	{
		BYTE* src = finaldata + dwWidth * (dwHeight-1);
		BYTE* dst = (BYTE *)desc.lpSurface;

		for (DWORD y=0; y<dwHeight; y++)
		{
			memcpy( dst, src, dwWidth );
			dst += desc.lPitch;
			src -= dwWidth;
		}
		avisurf->Unlock( 0 );
	}

	return TRUE;
}

void AviPlayWin::RestoreSurfaces()
{
	avisurf->Restore();
}

int AviPlayWin::SelectInitialDisplayMode()
{
	GetSystemPalette();
	
	int i, nummodes=GetNumDisplayModes();
	DWORD w,h,d;

	for (i=0;i<nummodes;i++)
	{
		DisplayModeDescription desc;
		GetDisplayModeDimensions( i, w, h, d );
		if (d==8)
		{
			desc.w=w;
			desc.h=h;
			desc.d=d;
			desc.desc.Format( "%dx%d", w, h );
			displaymode.Add( desc );
		}
	}

	int curdepth=GetDisplayDepth();
	if (curdepth!=8)
		ddraw2->SetDisplayMode( 640, 480, curdepth, 0, 0 );

	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==640 && h==480 && d==8)
			return i;
	}

	return 1;
}

void AviPlayWin::OnKeyDown(UINT key, UINT nRepCnt, UINT nFlags) 
{
	switch (key)
	{
		case VK_ESCAPE:
		case VK_SPACE:
		case VK_RETURN:
			ShowDialog();
			break;
	}
	
	DirectDrawWin::OnKeyDown(key, nRepCnt, nFlags);
}

void AviPlayWin::ShowDialog()
{
	const CRect& displayrect=GetDisplayRect();
	if (displayrect.Width()<640 || displayrect.Height()<480)
		ddraw2->SetDisplayMode( 640, 480, 8, 0, 0 );

	ClearSurface( backsurf, 0 );
	ClearSurface( primsurf, 0 );
	primsurf->SetPalette( syspal );
	
	ddraw2->FlipToGDISurface();
	ShowCursor( TRUE );

	if (avidialog==0)
	{
		avidialog=new AviDialog();
		avidialog->SetArray( &displaymode );
	}

	if (avistream)
		AVIStreamRelease( avistream ), avistream=0;

	if (avidialog->DoModal()==IDCANCEL)
	{
		PostMessage( WM_CLOSE );
		return;
	}
	ShowCursor( FALSE );

	fullfilename=avidialog->fullfilename;
	filename=avidialog->filename;
	pathname=avidialog->pathname;

	int index=avidialog->GetIndex();
	DWORD w,h,d;

	w=displaymode[index].w;
	h=displaymode[index].h;
	d=displaymode[index].d;
	ActivateDisplayMode( GetDisplayModeIndex( w, h, d ) );

	LoadAvi();
	CreateAviSurface();
	InstallPalette();
	curframe=startframe;
}

BOOL AviPlayWin::LoadAvi()
{
	long r;
	CWaitCursor cur;

	if (avistream)
		AVIStreamRelease( avistream ), avistream=0;
    r=AVIStreamOpenFromFile( &avistream, filename,
		streamtypeVIDEO, 0, OF_READ | OF_SHARE_EXCLUSIVE, 0 );
	TRACE("AVIStreamOpenFromFile: %s\n", r==0 ? "OK" : "failed" );

    r=AVIStreamFormatSize( avistream, 0, &fmtlen );
	TRACE("AVIStreamFormatSize: %s\n", r==0 ? "OK" : "failed" );

	int formatsize=fmtlen+sizeof(RGBQUAD)*256;
    if (srcfmt)
		delete [] srcfmt;
	srcfmt = (LPBITMAPINFOHEADER)new BYTE[formatsize];
	ZeroMemory( srcfmt, formatsize );

    if (dstfmt)
		delete [] dstfmt;
	dstfmt = (LPBITMAPINFOHEADER)new BYTE[formatsize];
	ZeroMemory( dstfmt, formatsize );

    r=AVIStreamReadFormat( avistream, 0, srcfmt, &fmtlen );
	TRACE("AVIStreamReadFormat: %s\n", r==0 ? "OK" : "failed" );

	TRACE(" --- %s ---\n", filename);
	TRACE(" biSize: %d\n", srcfmt->biSize );
	TRACE(" biWidth x biHeight: %dx%d\n", srcfmt->biWidth, srcfmt->biHeight ); 
	if (srcfmt->biPlanes != 1)
		TRACE(" - biPlanes: %d\n", srcfmt->biPlanes ); 
	TRACE(" biBitCount: %d\n", srcfmt->biBitCount );

	CString comp;
	switch (srcfmt->biCompression)
	{
		case BI_RGB:
			comp="BI_RGB";
			break;
		case BI_RLE8:
			comp="BI_RLE8";
			break;
		case BI_RLE4:
			comp="BI_RLE4";
			break;
		case BI_BITFIELDS:
			comp="BI_BITFIELDS";
			break;
	}
	TRACE(" biCompression: %s\n", comp ); 

	TRACE(" biSizeImage: %d\n", srcfmt->biSizeImage ); 
	TRACE(" ------------------\n");

    memcpy( dstfmt, srcfmt, fmtlen );
    dstfmt->biBitCount = 8;
    dstfmt->biCompression = BI_RGB;
	dstfmt->biSizeImage = dstfmt->biWidth * dstfmt->biHeight;

	startframe = AVIStreamStart( avistream );
	TRACE("stream start: %d\n", startframe);

    endframe = AVIStreamEnd( avistream );
	TRACE("stream end: %d\n", endframe );


	r=AVIStreamInfo( avistream, &streaminfo, sizeof(streaminfo) );
	TRACE("AVIStreamInfo: %s\n", r==0 ? "OK" : "failed" );

	buflen = dstfmt->biSizeImage;
	int finalbuflen=((dstfmt->biWidth+3) & ~3) * dstfmt->biHeight;

    if (streaminfo.dwSuggestedBufferSize)
        if ((LONG)streaminfo.dwSuggestedBufferSize < buflen)
		{
			TRACE("adjusting buflen to suggested size\n");
            buflen = (LONG)streaminfo.dwSuggestedBufferSize;
		}

	if (decomp)
		ICClose( decomp );
    decomp = ICDecompressOpen( ICTYPE_VIDEO, streaminfo.fccHandler, srcfmt, dstfmt );
	TRACE("ICDecompressOpen: %s\n", decomp ? "OK" : "failed");

    if (rawdata)
	{
		TRACE("delete [] rawdata...\n");
		delete [] rawdata;
	}
	rawdata = new BYTE[buflen];

	if (finaldata)
	{
		TRACE("delete [] finaldata...\n");
		delete [] finaldata;
	}
	finaldata = new BYTE[finalbuflen];

	return TRUE;
}

BOOL AviPlayWin::InstallPalette()
{
	ICDecompressGetPalette( decomp, srcfmt, dstfmt );

	PALETTEENTRY pe[256];
	LPBITMAPINFO info=(LPBITMAPINFO)dstfmt;
	for (int i=0; i<256; i++)
	{
		pe[i].peRed   = info->bmiColors[i].rgbRed;
		pe[i].peGreen = info->bmiColors[i].rgbGreen;
		pe[i].peBlue  = info->bmiColors[i].rgbBlue;
		pe[i].peFlags = 0;
	}
	
	if (avipal)
		avipal->Release();
    ddraw2->CreatePalette( DDPCAPS_8BIT, pe, &avipal, 0 );

	primsurf->SetPalette( avipal );

	return TRUE;
}

void AviPlayWin::OnRButtonDown(UINT nFlags, CPoint point) 
{
	ShowDialog();
	
	DirectDrawWin::OnRButtonDown(nFlags, point);
}

void AviPlayWin::OnDestroy() 
{
	DirectDrawWin::OnDestroy();

	if (avistream)
		AVIStreamRelease( avistream ), avistream=0;

	if (decomp)
		ICClose( decomp ), decomp=0;
	
	if (srcfmt)
		delete [] srcfmt, srcfmt=0;
	if (dstfmt)
		delete [] dstfmt, dstfmt=0;

	if (rawdata)
	{
		TRACE("delete [] rawdata...\n");
		delete [] rawdata, rawdata=0;
	}
	if (finaldata)
	{
		TRACE("delete [] finaldata..\n");
		delete [] finaldata, finaldata=0;;
	}

	if (avidialog)
		delete avidialog, avidialog=0;

	AVIFileExit();
}

void AviPlayWin::GetSystemPalette()
{
	PALETTEENTRY pe[256];
	HDC dc = ::GetDC( 0 );
	if (GetDeviceCaps(dc, RASTERCAPS) & RC_PALETTE)
	{
		GetSystemPaletteEntries( dc, 0, 256, pe );

		ddraw2->CreatePalette( DDPCAPS_8BIT, pe, &syspal, 0 );
	}
	::ReleaseDC( 0, dc );
}

int AviPlayWin::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (DirectDrawWin::OnCreate(lpCreateStruct) == -1)
		return -1;
	
    AVIFileInit();

	ShowDialog();
	
	return 0;
}
