// BmpDialog.cpp : implementation file
//

#include "headers.h"
#include "resource.h"
#include "BmpDialog.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// BmpDialog dialog



BmpDialog::BmpDialog(CWnd* pParent /*=NULL*/)
	: CDialog(BmpDialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(BmpDialog)
	filename = _T("");
	pathname = _T("");
	bmpspecs = _T("");
	//}}AFX_DATA_INIT

	index=-1;
}


void BmpDialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(BmpDialog)
	DDX_Control(pDX, IDC_MODELIST, modelist);
	DDX_Text(pDX, IDC_FILENAME, filename);
	DDX_Text(pDX, IDC_PATH, pathname);
	DDX_Text(pDX, IDC_BMPSPECS, bmpspecs);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(BmpDialog, CDialog)
	//{{AFX_MSG_MAP(BmpDialog)
	ON_BN_CLICKED(IDC_OPENBMP, OnOpenbmp)
	ON_LBN_SELCHANGE(IDC_MODELIST, OnSelchangeModelist)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

void BmpDialog::SetArrays( DisplayModeArray* p, DisplayModeArray* n)
{
	palettized=p;
	nonpalettized=n;
}

/////////////////////////////////////////////////////////////////////////////
// BmpDialog message handlers

void BmpDialog::OnOpenbmp() 
{
	static char BASED_CODE filter[] = "BMP Files (*.bmp)|*.bmp||";
	CFileDialog opendialog( TRUE, 0, 0, OFN_FILEMUSTEXIST, filter, this );
	if ( opendialog.DoModal() == IDOK )
	{
		filename = opendialog.GetFileName();
		fullfilename=opendialog.GetPathName();
		int pathlen=fullfilename.ReverseFind('\\');
		pathname=fullfilename.Left(pathlen);

		InitControls();
		UpdateData(FALSE);
	}

}

void BmpDialog::OnSelchangeModelist() 
{
	int cursel=modelist.GetCurSel();
	if (cursel==LB_ERR)
		return;

	GetDlgItem(IDOK)->EnableWindow();
}

void BmpDialog::OnOK() 
{
	index=modelist.GetCurSel();
	if (index==LB_ERR)
		index=-1;
	
	CDialog::OnOK();
}

BOOL BmpDialog::GetBmpDimensions( LPCTSTR filename, int* w, int* h, int* d)
{
	ifstream bmp( filename, ios::binary | ios::nocreate );
	if (!bmp.is_open())
	{
		TRACE("GetBmpDimensions: cannot open file '%s'\n", filename);
		return FALSE;
	}

	BITMAPFILEHEADER bmpfilehdr;
	bmp.read( (char*)&bmpfilehdr, sizeof(bmpfilehdr) );

	char* ptr=(char*)&bmpfilehdr.bfType;
	if (*ptr!='B' || *++ptr!='M')
	{
		TRACE("GetBmpDimensions: invalid bitmap\n");
		return FALSE;
	}

	BITMAPINFOHEADER bmpinfohdr;
	bmp.read( (char*)&bmpinfohdr, sizeof(bmpinfohdr) );

	if (w)
		*w=bmpinfohdr.biWidth;
	if (h)
		*h=bmpinfohdr.biHeight;
	if (d)
		*d=bmpinfohdr.biBitCount;

	return TRUE;
}

BOOL BmpDialog::OnInitDialog() 
{
	CDialog::OnInitDialog();

	InitControls();
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

BOOL BmpDialog::InitControls()
{
	GetDlgItem(IDOK)->EnableWindow( FALSE );
	
	int w, h, depth;
	if (GetBmpDimensions( fullfilename, &w, &h, &depth )==FALSE)
		return TRUE;

	bmpspecs.Format("%dx%dx%d", w, h, depth );

	if (depth<=8)
	{
		if (filepalettized==FALSE)
			index=-1;
		filepalettized=TRUE;
		modelist.ResetContent();
		int size=palettized->GetSize();
		for (int i=0;i<size;i++)
			modelist.InsertString( i, palettized->GetAt(i).desc );
	}
	else
	{
		if (filepalettized==TRUE)
			index=-1;
		filepalettized=FALSE;
		modelist.ResetContent();
		int size=nonpalettized->GetSize();
		for (int i=0;i<size;i++)
			modelist.InsertString( i, nonpalettized->GetAt(i).desc );
	}

	if (index!=-1 && modelist.GetCount()>0)
	{
		modelist.SetCurSel( index );
		GetDlgItem(IDOK)->EnableWindow( TRUE );
	}

	return TRUE;
}

