#include "Headers.h"
#include "resource.h"
#include "DriverDialog.h"
#include "BmpDialog.h"
#include "BmpViewWin.h"

#pragma comment (lib,"ddraw.lib")
#pragma comment (lib,"dxguid.lib")


BEGIN_MESSAGE_MAP(BmpViewWin, DirectDrawWin)
	//{{AFX_MSG_MAP(BmpViewWin)
	ON_WM_KEYDOWN()
	ON_WM_RBUTTONDOWN()
	ON_WM_DESTROY()
	ON_WM_CREATE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


BmpViewWin::BmpViewWin()
{
	syspal=0;

	bmpsurf=0;
	bmpdialog=0;
	x=0;
	y=0;

	update_screen=TRUE;
}

void BmpViewWin::DrawScene()
{
	if (update_screen && bmpsurf)
	{
		ClearSurface( backsurf, 0 );
		BltSurface( backsurf, bmpsurf, x, y );
		primsurf->Flip( 0, DDFLIP_WAIT );

		update_screen=FALSE;
	}
}

void BmpViewWin::RestoreSurfaces()
{/*
	CString filename;

	if (GetCurDisplayDepth()==8)
		filename="tri08.bmp";
	else
		filename="tri24.bmp";

	bmpsurf->Restore();
	LoadSurface( bmpsurf, filename );*/
}

int BmpViewWin::SelectInitialDisplayMode()
{
	DisplayModeDescription desc;
	int i, nummodes=GetNumDisplayModes();
	DWORD w,h,d;

	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		desc.w=w;
		desc.h=h;
		desc.d=d;
		desc.desc.Format("%dx%dx%d", w, h, d );

		if ( d==8 )
			palettemode.Add( desc );
		else
			nonpalettemode.Add( desc );
	}
	
	DWORD curdepth=GetDisplayDepth();

	//----- look for 640x480 with current pixel depth
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==640 && h==480 && d==curdepth)
			return i;
	}

	//----- barring that, try anything with the current depth
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (d==curdepth)
			return i;
	}

	//----- if that fails, try finding any 640x480 mode
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==640 && h==480)
			return i;
	}

	GetSystemPalette();
	
	return 0;
}

void BmpViewWin::OnKeyDown(UINT key, UINT nRepCnt, UINT nFlags) 
{
	switch (key)
	{
		case VK_UP:
			Up();
			break;
		case VK_DOWN:
			Down();
			break;
		case VK_LEFT:
			Left();
			break;
		case VK_RIGHT:
			Right();
			break;
		case VK_HOME:
			Home();
			break;
		case VK_END:
			End();
			break;
		case VK_PRIOR:
			PageUp();
			break;
		case VK_NEXT:
			PageDown();
			break;
		case VK_ESCAPE:
		case VK_SPACE:
		case VK_RETURN:
			ShowDialog();
			break;
	}
	
	DirectDrawWin::OnKeyDown(key, nRepCnt, nFlags);
}

void BmpViewWin::ShowDialog()
{
	CRect displayrect=GetDisplayRect();
	if (displayrect.Width()<640)
		ddraw2->SetDisplayMode( 640, 480, 8, 0, 0 );

	if (GetDisplayDepth()==8)
	{
		ClearSurface( backsurf, 0 );
		primsurf->SetPalette( syspal );
	}
	else
	{
		BltSurface( backsurf, bmpsurf, x, y );
	}
	
	ddraw2->FlipToGDISurface();
	ShowCursor(TRUE);

	if (bmpdialog==0)
	{
		bmpdialog=new BmpDialog();
		bmpdialog->SetArrays( &palettemode, &nonpalettemode );
	}

	if (bmpdialog->DoModal()==IDCANCEL)
	{
		PostMessage( WM_CLOSE );
		return;
	}

	fullfilename=bmpdialog->fullfilename;
	filename=bmpdialog->filename;
	pathname=bmpdialog->pathname;

	int index=bmpdialog->GetIndex();
	DWORD w,h,d;
	if (bmpdialog->FilePalettized())
	{
		w=palettemode[index].w;
		h=palettemode[index].h;
		d=palettemode[index].d;
	}
	else
	{
		w=nonpalettemode[index].w;
		h=nonpalettemode[index].h;
		d=nonpalettemode[index].d;
	}

	if (GetDisplayDepth()==8)
		primsurf->SetPalette( palette );

	ActivateDisplayMode( GetDisplayModeIndex( w, h, d) );

	LoadBmp();
	
	ShowCursor(FALSE);
}

BOOL BmpViewWin::LoadBmp()
{
	CWaitCursor cur;

	LPDIRECTDRAWSURFACE surf;
	surf=CreateSurface( filename, TRUE );
	if (surf)
	{
		if (bmpsurf)
			bmpsurf->Release();
		bmpsurf=surf;
	}
	else
	{
		TRACE("failed to load new file\n");
		return FALSE;
	}

	displayrect=GetDisplayRect();
	TRACE("display: %d %d\n", displayrect.right, displayrect.bottom);
	GetSurfaceRect( bmpsurf, bmprect );
	TRACE("surface: %d %d\n", bmprect.right, bmprect.bottom);
	
	int mx = displayrect.Width()-bmprect.Width();
	if (mx<0)
	{
		xscroll=TRUE;
		xlimit=mx;
		x=0;
	}
	else
	{
		xscroll=FALSE;
		x=mx/2;
	}

	int my = displayrect.Height()-bmprect.Height();
	if (my<0)
	{
		yscroll=TRUE;
		ylimit=my;
		y=0;
	}
	else
	{
		yscroll=FALSE;
		y=my/2;
	}

	update_screen=TRUE;

	return TRUE;
}

void BmpViewWin::PageUp()
{
	if (yscroll)
	{
		if (y-displayrect.Height()>0)
		{
			y-=displayrect.Height();
			update_screen=TRUE;
		}
		else 
		{
			y=0;
			update_screen=TRUE;
		}
	}
}

void BmpViewWin::PageDown()
{
	if (yscroll)
	{
		if (y+displayrect.Height()<=ylimit)
		{
			y+=displayrect.Height();
			update_screen=TRUE;
		}
		else 
		{
			y=ylimit;
			update_screen=TRUE;
		}
	}
}

void BmpViewWin::Home()
{
	if (xscroll && x!=0)
	{
		x=0;
		update_screen=TRUE;
	}
	if (yscroll && y!=0)
	{
		y=0;
		update_screen=TRUE;
	}
}

void BmpViewWin::End()
{
	if (yscroll)
	{
		y=-(bmprect.Height()-displayrect.Height());
		update_screen=TRUE;
	}
	if (xscroll)
	{
		x=-(bmprect.Width()-displayrect.Width());
		update_screen=TRUE;
	}
}

void BmpViewWin::Right(int inc)
{
	if (!xscroll)
		return;

	if (x-inc>=xlimit)
	{
		x-=inc;
		update_screen=TRUE;
	}
	else
	{
		x=xlimit;
		update_screen=TRUE;
	}
}

void BmpViewWin::Left(int inc)
{
	if (!xscroll)
		return;

	if (x+inc<0)
	{
		x+=inc;
		update_screen=TRUE;
	}
	else
	{
		x=0;
		update_screen=TRUE;
	}
}

void BmpViewWin::Down(int inc)
{
	if (!yscroll)
		return;

	if (y-inc>=ylimit)
	{
		y-=inc;
		update_screen=TRUE;
	}
	else
	{
		y=ylimit;
		update_screen=TRUE;
	}
}

void BmpViewWin::Up(int inc)
{
	if (!yscroll)
		return;

	if (y+inc<0)
	{
		y+=inc;
		update_screen=TRUE;
	}
	else
	{
		y=0;
		update_screen=TRUE;
	}
}

void BmpViewWin::OnRButtonDown(UINT nFlags, CPoint point) 
{
	ShowDialog();
	
	DirectDrawWin::OnRButtonDown(nFlags, point);
}

void BmpViewWin::OnDestroy() 
{
	DirectDrawWin::OnDestroy();
	
	if (bmpdialog)
		delete bmpdialog, bmpdialog=0;
}

void BmpViewWin::GetSystemPalette()
{
    PALETTEENTRY pe[256];
    HDC dc = ::GetDC( 0 );
    if (GetDeviceCaps(dc, RASTERCAPS) & RC_PALETTE)
    {
        GetSystemPaletteEntries( dc, 0, 256, pe );

        ddraw2->CreatePalette( DDPCAPS_8BIT, pe, &syspal, 0 );
    }
    ::ReleaseDC( 0, dc );
}

int BmpViewWin::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (DirectDrawWin::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	ShowDialog();
	
	return 0;
}
