#include "Headers.h"
#include "resource.h"
#include "DriverDialog.h"
#include "BumperWin.h"
#include "Sprite.h"

#pragma comment (lib,"winmm.lib")
#pragma comment (lib,"ddraw.lib")
#pragma comment (lib,"dxguid.lib")


const DWORD desiredwidth=640;
const DWORD desiredheight=480;
const DWORD desireddepth=8;


BEGIN_MESSAGE_MAP(BumperWin, DirectDrawWin)
	//{{AFX_MSG_MAP(BumperWin)
	ON_WM_KEYDOWN()
	ON_WM_DESTROY()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


BumperWin::BumperWin()
{
	nsprites=0;
}

BOOL BumperWin::CreateCustomSurfaces()
{
	DDCOLORKEY  ddck;
	ddck.dwColorSpaceLowValue = 0;
	ddck.dwColorSpaceHighValue = 0;

	LPDIRECTDRAWSURFACE surf;

	srand( time(0) );

	CString msg="Can't find ";

	surf=CreateSurface( "diamond.bmp", TRUE );
	if (surf==0)
	{
		msg+="diamond.bmp";
		Fatal( msg );
	}
	surf->SetColorKey( DDCKEY_SRCBLT, &ddck );
	sprite[nsprites++]=new Sprite( surf, 0, 0 );
	sprite[nsprites++]=new Sprite( surf, 150, 0 );

	surf=CreateSurface( "triangle.bmp" );
	if (surf==0)
	{
		msg+="triangle.bmp";
		Fatal( msg );
	}
	surf->SetColorKey( DDCKEY_SRCBLT, &ddck );
	sprite[nsprites++]=new Sprite( surf, 0, 150 );
	sprite[nsprites++]=new Sprite( surf, 150, 150 );

	surf=CreateSurface( "rect.bmp" );	
	if (surf==0)
	{
		msg+="rect.bmp";
		Fatal( msg );
	}
	surf->SetColorKey( DDCKEY_SRCBLT, &ddck );
	sprite[nsprites++]=new Sprite( surf, 0, 300 );
	sprite[nsprites++]=new Sprite( surf, 150, 300 );


	surf=CreateSurface( "oval.bmp" );	
	if (surf==0)
	{
		msg+="oval.bmp";
		Fatal( msg );
	}
	surf->SetColorKey( DDCKEY_SRCBLT, &ddck );
	sprite[nsprites++]=new Sprite( surf, 300, 0 );
	sprite[nsprites++]=new Sprite( surf, 300, 150 );

	text=CreateSurface("text.bmp");	
	if (text==0)
	{
		msg+="text.bmp";
		Fatal( msg );
	}
	text->SetColorKey( DDCKEY_SRCBLT, &ddck );

	return TRUE;
}

void BumperWin::DrawScene()
{
	ASSERT(nsprites>0);
	ASSERT(text);

	for (int s1=0;s1<nsprites;s1++)
		for (int s2=s1+1;s2<nsprites;s2++)
			if (SpritesCollide( sprite[s1], sprite[s2] ))
			{
				sprite[s1]->Hit( sprite[s2] );
				sprite[s2]->Hit( sprite[s1] );
			}

	for (int i=0;i<nsprites;i++)
		sprite[i]->Update();

	ClearSurface( backsurf, 0 );
	for (i=0;i<nsprites;i++)
	{
		Sprite* s=sprite[i];
		BltSurface( backsurf, *s, s->GetX(), s->GetY(), TRUE );
	}

	BltSurface( backsurf, text, 0, 448, TRUE );

	primsurf->Flip( 0, DDFLIP_WAIT );
}

void BumperWin::RestoreSurfaces()
{
	for (int i=0;i<nsprites;i++)
		sprite[i]->GetSurf()->Restore();

	LoadSurface( *sprite[0], "diamond.bmp");
	LoadSurface( *sprite[1], "diamond.bmp");

	LoadSurface( *sprite[2], "triangle.bmp" );
	LoadSurface( *sprite[3], "triangle.bmp" );

	LoadSurface( *sprite[4], "rect.bmp" );
	LoadSurface( *sprite[5], "rect.bmp" );

	LoadSurface( *sprite[6], "oval.bmp" );
	LoadSurface( *sprite[7], "oval.bmp" );

	text->Restore();
	LoadSurface( text, "text.bmp" );
}

int BumperWin::SelectDriver()
{
	int numdrivers=GetNumDrivers();
	if (numdrivers==1)
		return 0;

	CArray<CString, CString> drivers;
	for (int i=0;i<numdrivers;i++)
	{
		LPSTR desc, name;
		GetDriverInfo( i, 0, &desc, &name );
		drivers.Add(desc);
	}

	DriverDialog dialog;
	dialog.SetContents( &drivers );
	if (dialog.DoModal()!=IDOK)
		return -1;

	return dialog.GetSelection();
}

int BumperWin::SelectInitialDisplayMode()
{
	DWORD curdepth=GetDisplayDepth();
	int i, nummodes=GetNumDisplayModes();
	DWORD w,h,d;

	if (curdepth!=desireddepth)
		ddraw2->SetDisplayMode( 640, 480, curdepth, 0, 0 );

	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==desiredwidth && h==desiredheight && d==desireddepth)
			return i;
	}

	ddraw2->RestoreDisplayMode();
	ddraw2->Release(), ddraw2=0;
	AfxMessageBox("Can't find 8-bit mode on this device");

	return -1;
}

void BumperWin::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	switch (nChar)
	{
	case VK_ESCAPE:
		PostMessage( WM_CLOSE );
		break;
	case VK_SPACE:
	case VK_RETURN:
		for (int i=0;i<nsprites;i++)
			sprite[i]->CalcVector();
		break;
	}
	
	DirectDrawWin::OnKeyDown(nChar, nRepCnt, nFlags);
}

BOOL BumperWin::SpritesCollide(Sprite* sprite1, Sprite* sprite2)
{
	ASSERT(sprite1 && sprite2);

	if (SpritesCollideRect( sprite1, sprite2 ))
		if (SpritesCollidePixel( sprite1, sprite2 ))
			return TRUE;

	return FALSE;
}

BOOL BumperWin::SpritesCollideRect(Sprite* sprite1, Sprite* sprite2)
{
	CRect rect1 = sprite1->GetRect();
	CRect rect2 = sprite2->GetRect();
	CRect r = rect1 & rect2;

	// if all fields are zero, then rects don't overlap
	return !(r.left==0 && r.top==0 && r.right==0 && r.bottom==0);
}

BOOL BumperWin::SpritesCollidePixel(Sprite* sprite1, Sprite* sprite2)
{
	// get locationgs of both sprites
	CRect rect1=sprite1->GetRect();
	CRect rect2=sprite2->GetRect();

	// calc intersecting area of the two rects
	CRect irect = rect1 & rect2;

	// we're assuming that the rects to indeed intersect...
	ASSERT(!(irect.left==0 && irect.top==0 && 
		irect.right==0 && irect.bottom==0));

	// calc target areas for each sprite
	CRect r1target = rect1 & irect;
	r1target.OffsetRect( -rect1.left, -rect1.top );
	r1target.right--;
	r1target.bottom--;

	CRect r2target = rect2 & irect;
	r2target.OffsetRect( -rect2.left, -rect2.top );
	r2target.right--;
	r2target.bottom--;

	// store dims of target area
	int width=irect.Width();
	int height=irect.Height();

	// prepare description structs for both surfaces
	DDSURFACEDESC desc1, desc2;
	ZeroMemory( &desc1, sizeof(desc1) );
	ZeroMemory( &desc2, sizeof(desc2) );
	desc1.dwSize = sizeof(desc1);
	desc2.dwSize = sizeof(desc2);

	BYTE* surfptr1;  // used to point to start of surface memory
	BYTE* surfptr2;  
	BYTE* pixel1;    // used to point to specific pixels within
	BYTE* pixel2;    // surface memory
	BOOL ret=FALSE;

	// extract DirectDraw surface pointers
	LPDIRECTDRAWSURFACE surf1=sprite1->GetSurf();
	LPDIRECTDRAWSURFACE surf2=sprite2->GetSurf();

	// are both sprites using the same surface?
	if (surf1==surf2)
	{
		surf1->Lock( 0, &desc1, DDLOCK_WAIT, 0 );
		surfptr1=(BYTE*)desc1.lpSurface;

		for (int yy=0;yy<height;yy++)
		{
			for (int xx=0;xx<width;xx++)
			{
				pixel1=surfptr1+(yy+r1target.top)*desc1.lPitch
						+(xx+r1target.left);
				pixel2=surfptr1+(yy+r2target.top)*desc1.lPitch
						+(xx+r2target.left);
				if (*pixel1 && *pixel2)
				{
					ret=TRUE;
					goto done_same_surf;
				}
			}
		}

		done_same_surf:
		surf1->Unlock( surfptr1 );
		return ret;
	}

	// each sprite is represented by a different surface,
	// lock both and check for collision
	surf1->Lock( 0, &desc1, DDLOCK_WAIT, 0 );
	surfptr1=(BYTE*)desc1.lpSurface;

	surf2->Lock( 0, &desc2, DDLOCK_WAIT, 0 );
	surfptr2=(BYTE*)desc2.lpSurface;

	for (int yy=0;yy<height;yy++)
	{
		for (int xx=0;xx<width;xx++)
		{
			pixel1=surfptr1+(yy+r1target.top)*desc1.lPitch
					+(xx+r1target.left);
			pixel2=surfptr2+(yy+r2target.top)*desc2.lPitch
					+(xx+r2target.left);
			if (*pixel1 && *pixel2)
			{
				ret=TRUE;
				goto done;
			}
		}
	}
	
	done:
	surf2->Unlock( surfptr2 );
	surf1->Unlock( surfptr1 );

	return ret;
}

void BumperWin::OnDestroy() 
{
	DirectDrawWin::OnDestroy();
	
	for (int i=0;i<nsprites;i++)
		delete sprite[i];
}
