#include "Headers.h"
#include "resource.h"
#include "DriverDialog.h"
#include "CursorWin.h"

#pragma comment (lib,"ddraw.lib")
#pragma comment (lib,"dinput.lib")
#pragma comment (lib,"dxguid.lib")


LPDIRECTINPUTDEVICE CursorWin::mouse;

CCriticalSection CursorWin::critsection;
CWinThread* CursorWin::mousethread;
CEvent* CursorWin::mouse_event[2];

int CursorWin::cursor_width;
int CursorWin::cursor_height;
LPDIRECTDRAWSURFACE CursorWin::cursor;
LPDIRECTDRAWSURFACE CursorWin::cursor_under;
LPDIRECTDRAWSURFACE CursorWin::cursor_union;

int CursorWin::curx, CursorWin::cury;
int CursorWin::oldcurx, CursorWin::oldcury;
CList<MouseClickData, MouseClickData> CursorWin::mouseclickqueue;

const COLORREF textcolor=RGB(250, 200, 200);
const COLORREF textshadow=RGB(150, 100, 100);
//const COLORREF menubarcolor=RGB(100, 75, 200);

const int screen_width=640;
const int screen_height=480;

const int delay_value[]={ 0, 10, 50, 100, 250, 500 };
const int dm_width=100;
const int dm_header=35;
const int dm_texty=65;
const int dm_entrysize=20;
const int dm_margin=20;
const int dm_height=sizeof(delay_value)/sizeof(int)*dm_entrysize+40;

const int coilx=160;
const int coily=120;


BEGIN_MESSAGE_MAP(CursorWin, DirectDrawWin)
	//{{AFX_MSG_MAP(CursorWin)
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_ACTIVATE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


CursorWin::CursorWin()
{
	dinput=0;
	mouse=0;

	cursor=0;
	cursor_width=0;
	cursor_height=0;
	cursor_under=0;
	cursor_union=0;
	curx=320;
	cury=240;

	dm_surf=0;
	dm_index=0;

	largefont = CreateFont( 28, 0, 0, 0,
		FW_NORMAL, FALSE, FALSE, FALSE,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		VARIABLE_PITCH,
		"Arial" );

	smallfont = CreateFont( 18, 0, 0, 0,
		FW_NORMAL, FALSE, FALSE, FALSE,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		VARIABLE_PITCH,
		"Arial" );
}

BOOL CursorWin::CreateCustomSurfaces()
{
	int ext;
	if (GetDisplayDepth()==8)
	{
		ext=8;
		menubarfillcolor=2;
	}
	else
	{
		ext=24;
		menubarfillcolor=RGBtoPixel( 100, 75, 200 );
	}

	CString msg="Can't find ";

	char filename[20];
	sprintf( filename, "cursor_%02d.bmp", ext );
	cursor=CreateSurface( filename, TRUE );
	if (cursor==0)
		Fatal(msg+filename);

	GetSurfaceDimensions( cursor, (DWORD&)cursor_width, (DWORD&)cursor_height );
	dm_surf=CreateSurface( dm_width, dm_height );
	UpdateDelaySurface();
	
	cursor_under=CreateSurface( cursor_width, cursor_height );
	cursor_union=CreateSurface( cursor_width*2, cursor_height*2 );

	DDCOLORKEY  colorkey;
	colorkey.dwColorSpaceLowValue = 0;
	colorkey.dwColorSpaceHighValue = 0;
	cursor->SetColorKey( DDCKEY_SRCBLT, &colorkey );

	for (int i=0;i<coil_frames;i++)
	{
		sprintf( filename, "stwist%02d_%02d.bmp", i, ext );
		coil[i]=CreateSurface( filename );
		if (coil[i]==0)
		{
			CString str="Cannot load file: ";
			Fatal(str+=filename);
		}
	}

	return TRUE;
}

void CursorWin::DrawScene()
{
	//------ check for ESCAPE key -------
	static char key[256];
	keyboard->GetDeviceState( sizeof(key), &key );
	if ( key[DIK_ESCAPE] & 0x80 )
		PostMessage( WM_CLOSE );

	if ( key[DIK_S] & 0x80 )
	{
		static BOOL already;
		if (!already)
			SaveSurface( primsurf, "out.bmp" );
	}


	//------ do regular app stuff ------
	ClearSurface( backsurf, 0 );

	BltSurface( backsurf, dm_surf, 539, 0 );

	static coil_idx;
	BltSurface( backsurf, coil[coil_idx], coilx, coily );
	coil_idx=(coil_idx+1)%coil_frames;

	//------ do synchronized stuff ------
	critsection.Lock();

	//------ save back buffer area under cursor
	RECT src;
	src.left=curx;
	src.top=cury;
	src.right=curx+cursor_width;
	src.bottom=cury+cursor_height;
	cursor_under->BltFast( 0, 0, backsurf, &src, DDBLTFAST_WAIT );

	//------ draw cursor to back buffer
	backsurf->BltFast( curx, cury, cursor, 0, DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT );

	primsurf->Flip( 0, DDFLIP_WAIT );
	while (primsurf->GetFlipStatus(DDGFS_ISFLIPDONE)!=DD_OK)
		;  // do nothing (wait for flip to complete)

	int x, y;
	BOOL newclick=FALSE;
	int count=mouseclickqueue.GetCount();
	while (count--)
	{
		MouseClickData mc=mouseclickqueue.RemoveTail();
		if (mc.button==0)
		{
			x=mc.x;
			y=mc.y;
			newclick=TRUE;
		}
	}

	critsection.Unlock();
	//------ end of synchronized stuff -------

	//------ sleep according to delay menu ----
	if ( delay_value[dm_index]!=0)
		Sleep( delay_value[dm_index] );

	//------ update the delay menu --------
	if (newclick)
	{
		int max_index=sizeof(delay_value)/sizeof(int)-1;
		int menux=screen_width-dm_width+dm_margin;
		int menuw=dm_width-dm_margin*2;
		if (x>=menux && x<=menux+menuw)
		{
			int index=(y-dm_header)/dm_entrysize;
			if (index>=0 && index<=max_index && index!=dm_index)
			{
				dm_index=index;
				UpdateDelaySurface();
			}
		}
	}
}

BOOL CursorWin::InitKeyboard()
{
	HRESULT r;
	r = dinput->CreateDevice( GUID_SysKeyboard, &keyboard, 0 ); 
	if (r!=DI_OK)
	{
		TRACE("CreateDevice(keyboard) failed");	
		return FALSE;
	}
	
	r = keyboard->SetDataFormat( &c_dfDIKeyboard );
	if (r!=DI_OK)
	{
		TRACE("keyboard->SetDataFormat() failed\n");
		return FALSE;
	}

	r=keyboard->SetCooperativeLevel( GetSafeHwnd(), 
			DISCL_FOREGROUND | DISCL_NONEXCLUSIVE); 
	if (r!=DI_OK)
	{
		TRACE("keyboard->SetCooperativeLevel() failed\n");
		return FALSE;
	}

	return TRUE;
}

BOOL CursorWin::InitMouse()
{
	HRESULT r;
	r = dinput->CreateDevice( GUID_SysMouse, &mouse, 0 ); 
	if (r!=DI_OK)
	{
		TRACE("CreateDevice(mouse) failed\n");	
		return FALSE;
	}

	r = mouse->SetDataFormat( &c_dfDIMouse );
	if (r!=DI_OK)
	{
		TRACE("mouse->SetDataFormat() failed\n");
		return FALSE;
	}

	r = mouse->SetCooperativeLevel( GetSafeHwnd(),
			DISCL_NONEXCLUSIVE | DISCL_FOREGROUND );
	if (r!=DI_OK)
	{
		TRACE("mouse->SetCooperativeLevel() failed\n");
		return FALSE;
	}

	DIPROPDWORD property;
	property.diph.dwSize=sizeof(DIPROPDWORD);
	property.diph.dwHeaderSize=sizeof(DIPROPHEADER);
	property.diph.dwObj=0;
	property.diph.dwHow=DIPH_DEVICE;
	property.dwData=64;

	r = mouse->SetProperty( DIPROP_BUFFERSIZE, &property.diph );
	if (r!=DI_OK)
	{
		TRACE("mouse->SetProperty() failed (buffersize)\n");
		return FALSE;
	}

	mouse_event[mouse_event_index]=new CEvent;
	mouse_event[quit_event_index]=new CEvent;

	r = mouse->SetEventNotification( *mouse_event[mouse_event_index] );
	if (r!=DI_OK)
	{
		TRACE("mouse->SetEventNotification() failed\n");
		return FALSE;
	}

	mousethread=AfxBeginThread( (AFX_THREADPROC)MouseThread, this,
			THREAD_PRIORITY_TIME_CRITICAL, 0, CREATE_SUSPENDED );

	return TRUE;
}

DWORD CursorWin::MouseThread(LPVOID p)
{
	TRACE("starting mouse thread\n" );

	CursorWin* win=(CursorWin*)p;

	while(TRUE)
	{
		CMultiLock mlock( (CSyncObject**)mouse_event, 2 );
		DWORD event=mlock.Lock( INFINITE, FALSE );
		if (event-WAIT_OBJECT_0==quit_event_index)
		{
			TRACE("got quit message: quitting mouse thread\n");
			return 0;
		}

		critsection.Lock();

		oldcurx=curx;
		oldcury=cury;
		BOOL buffer_empty=FALSE;
		while (!buffer_empty)
		{
			DIDEVICEOBJECTDATA data;
			DWORD elements=1;
			if (mouse==0)
			{
				TRACE("mouse pointer invalid: quitting mouse thread\n");
				return 0;
			}
			HRESULT r=mouse->GetDeviceData( sizeof(data), &data, &elements, 0 );
			if (r==DI_OK && elements==1)
			{
				static MouseClickData mc;
				switch(data.dwOfs)
				{
					case DIMOFS_X:
						curx+=data.dwData;
						break;
					case DIMOFS_Y: 
						cury+=data.dwData;
						break;
					case DIMOFS_BUTTON0:
						if (data.dwData & 0x80)
						{
							mc.x=curx;
							mc.y=cury;
							mc.button=0;
							mouseclickqueue.AddHead( mc );
						}
						break;
					case DIMOFS_BUTTON1:
						if (data.dwData & 0x80)
						{
							mc.x=curx;
							mc.y=cury;
							mc.button=1;
							mouseclickqueue.AddHead( mc );
						}
						break;
				}
			}
			else
				buffer_empty=TRUE;
		}

		if (curx<0)
			curx=0;
		if (cury<0)
			cury=0;
		if (curx>=screen_width-cursor_width)
			curx=screen_width-cursor_width-1;
		if (cury>=screen_height-cursor_height)
			cury=screen_height-cursor_height-1;

		if (curx==oldcurx && cury==oldcury)
		{
			//----- no cursor update is necesssary ------
			goto nevermind;
		}
		else if (abs(curx-oldcurx) >= cursor_width || abs(cury-oldcury) >= cursor_height)
		{
			//----- simple case: new and old cursor rects don't overlap -----
			win->UpdateCursorSimpleCase( curx, cury, oldcurx, oldcury );
		}
		else
		{
			//----- complex case: new and old cursor rects overlap -----
			win->UpdateCursorComplexCase( curx, cury, oldcurx, oldcury );
		}
		
		nevermind:;
		critsection.Unlock();
	}
	
	TRACE("leaving mouse thread\n");
	return 0;
};

void CursorWin::OnActivate(UINT nState, CWnd* pWndOther, BOOL bMinimized) 
{
	DirectDrawWin::OnActivate(nState, pWndOther, bMinimized);
	
	if (nState!=WA_INACTIVE)
	{
		if (mouse)
		{
			TRACE("mouse->Acquire()\n");
			mouse->Acquire();
		}
		if (keyboard)
		{
			TRACE("keyboard->Acquire()\n");
			keyboard->Acquire();
		}
	}
}

int CursorWin::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	HRESULT r=DirectInputCreate( AfxGetInstanceHandle(), 
			DIRECTINPUT_VERSION, &dinput, 0 ); 
	if (r!=DI_OK)
	{
		AfxMessageBox("DirectInputCreate() failed");
		return -1;
	}

	if (InitMouse()==FALSE)
		return -1;

	if (InitKeyboard()==FALSE)
		return -1;
	
	if (DirectDrawWin::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	mousethread->ResumeThread();

	return 0;
}

void CursorWin::OnDestroy() 
{
	critsection.Lock();

	DirectDrawWin::OnDestroy();

	if (mouse)
	{
		TRACE("mouse->Unacquire()\n");
		mouse->Unacquire();

		TRACE("sending mouse quit message...\n");
		mouse_event[quit_event_index]->SetEvent();

		Sleep(100);  // give mouse thread a chance to respond
		TRACE("Releasing mouse pointer...\n");
		mouse->Release(), mouse=0;

		delete mouse_event[mouse_event_index];
		delete mouse_event[quit_event_index];
	}

	if (keyboard)
		keyboard->Release(), keyboard=0;

	if (dinput)
		dinput->Release(), dinput=0;

	critsection.Unlock();
}

void CursorWin::RestoreSurfaces()
{

}

int CursorWin::SelectDriver()
{
	int numdrivers=GetNumDrivers();
	if (numdrivers==1)
		return 0;

	CArray<CString, CString> drivers;
	for (int i=0;i<numdrivers;i++)
	{
		LPSTR desc, name;
		GetDriverInfo( i, 0, &desc, &name );
		drivers.Add( desc );
	}

	DriverDialog dialog;
	dialog.SetContents( &drivers );
	if (dialog.DoModal()!=IDOK)
		return -1;

	return dialog.GetSelection();
}

int CursorWin::SelectInitialDisplayMode()
{
	DWORD curdepth=GetDisplayDepth();
	int i, nummodes=GetNumDisplayModes();
	DWORD w,h,d;

	//--- Some ddraw configurations fail if the initial
	//--- call to SetDisplayMode() tries to change the
	//--- current display depth. To get around this
	//--- we activate a mode that has the same depth as
	//--- the desktop, and then choose our display mode.
	if (curdepth!=8)
		ddraw2->SetDisplayMode( 640, 480, curdepth, 0, 0 );

	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==screen_width && h==screen_height && d==8)
			return i;
	}

	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==screen_width && h==screen_height && d==16)
			return i;
	}

	AfxMessageBox("Desired display mode not supported");
	return -1;
}

BOOL CursorWin::UpdateCursorComplexCase(int curx, int cury, int oldcurx, int oldcury)
{
	RECT src;
	HRESULT r;

	int unionx=min(curx, oldcurx);
	int uniony=min(cury, oldcury);
	int unionw=max(curx, oldcurx)-unionx+cursor_width;
	int unionh=max(cury, oldcury)-uniony+cursor_height;

	//----- Blt 1: copy the rect union to the scratch buffer --------
	src.left=unionx;
	src.top=uniony;
	src.right=unionx+unionw;
	src.bottom=uniony+unionh;

	r=cursor_union->BltFast( 0, 0, primsurf, &src, DDBLTFAST_WAIT );
	if (r!=DD_OK)
	{
		TRACE("Blt 1 failed\n");
		CheckResult(r);
	}

	//------ Blt 2: erase old cursor from scratch buffer ---------
	r=cursor_union->BltFast( oldcurx-unionx, oldcury-uniony, cursor_under, 0, DDBLTFAST_WAIT );
	if (r!=DD_OK)
	{
		TRACE("Blt 2 failed\n");
		CheckResult(r);
	}

	//------ Blt 3: save area under new cursor ---------
	src.left=curx-unionx;
	src.top=cury-uniony;
	src.right=src.left+cursor_width;
	src.bottom=src.top+cursor_height;

	r=cursor_under->BltFast( 0, 0, cursor_union, &src, DDBLTFAST_WAIT );
	if (r!=DD_OK)
	{
		TRACE("Blt 3 failed\n");
		CheckResult(r);
	}

	//------ Blt 4: draw new cursor into scratch buffer ---------
	r=cursor_union->BltFast( curx-unionx, cury-uniony, cursor, 0, DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT );
	if (r!=DD_OK)
	{
		TRACE("Blt 4 failed\n");
		CheckResult(r);
	}

	//------- Blt 5: blt scratch buffer back to primary buffer --------
	src.left=0;
	src.top=0;
	src.right=unionw;
	src.bottom=unionh;

	r=primsurf->BltFast( unionx, uniony, cursor_union, &src, DDBLTFAST_WAIT );
	if (r!=DD_OK)
	{
		TRACE("Blt 5 failed\n");
		CheckResult(r);
	}

	return TRUE;
}

BOOL CursorWin::UpdateCursorSimpleCase(int curx, int cury, int oldcurx, int oldcury)
{
	RECT src;
	HRESULT r;

	//------ Blt 1: erase old cursor ----------
	r=primsurf->BltFast( oldcurx, oldcury, cursor_under, 0, DDBLTFAST_WAIT );
	if (r!=DD_OK)
	{
		TRACE("Blt 1 failed\n");
		CheckResult(r);
	}

	//------ Blt 2: save area under new cursor -----------
	src.left=curx;
	src.top=cury;
	src.right=curx+cursor_width;
	src.bottom=cury+cursor_height;

	r=cursor_under->BltFast( 0, 0, primsurf, &src, DDBLTFAST_WAIT );
	if (r!=DD_OK)
	{
		TRACE("Blt 2 failed\n");
		CheckResult(r);
	}

	//------ Blt 3: draw new cursor ----------
	r=primsurf->BltFast( curx, cury, cursor, 0, DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT );
	if (r!=DD_OK)
	{
		TRACE("Blt 3 failed\n");
		CheckResult(r);
	}

	return TRUE;
}

BOOL CursorWin::UpdateDelaySurface()
{
	ClearSurface( dm_surf, 0 );

	RECT r;
	r.left=dm_margin;
	r.top=dm_header+dm_index*dm_entrysize;
	r.right=dm_width-dm_margin;
	r.bottom=r.top+18;

	DDBLTFX bltfx;
	ZeroMemory( &bltfx, sizeof(bltfx) );
	bltfx.dwSize = sizeof(bltfx);
	bltfx.dwFillColor = menubarfillcolor;

	dm_surf->Blt( &r, 0, 0, DDBLT_COLORFILL | DDBLT_WAIT, &bltfx );

	HDC hdc;
	dm_surf->GetDC( &hdc );

	//------- header text ---------
	SelectObject( hdc, largefont );
	SetTextAlign( hdc, TA_CENTER );
	SetBkMode( hdc, TRANSPARENT );

	SetTextColor( hdc, textshadow );
	TextOut( hdc, 50, 1, "Delay", 5 );

	SetTextColor( hdc, textcolor );
	TextOut( hdc, 50, 0, "Delay", 5 );

	//------- delay menu entries ------
	SelectObject( hdc, smallfont );
	SetTextAlign( hdc, TA_RIGHT );

	for (int i=0; i<sizeof(delay_value)/sizeof(int); i++)
	{
		char buf[10];
		int len=sprintf( buf, "%d", delay_value[i] );

		SetTextColor( hdc, textshadow );
		TextOut( hdc, dm_texty+1, (dm_header+1)+i*dm_entrysize, buf, len );

		SetTextColor( hdc, textcolor );
		TextOut( hdc, dm_texty, dm_header+i*dm_entrysize, buf, len );
	}

	dm_surf->ReleaseDC( hdc );

	return TRUE;
}

