#include "Headers.h"
#include "resource.h"
#include "DriverDialog.h"
#include "SmearWin.h"

#pragma comment (lib,"ddraw.lib")
#pragma comment (lib,"dinput.lib")
#pragma comment (lib,"dxguid.lib")


BEGIN_MESSAGE_MAP(SmearWin, DirectDrawWin)
	//{{AFX_MSG_MAP(SmearWin)
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_ACTIVATE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


SmearWin::SmearWin()
{
	dinput=0;
	mouse=0;
	keyboard=0;

	x=320;
	y=240;
	sphere=0;
}

BOOL SmearWin::CreateCustomSurfaces()
{
	if (GetCurDisplayDepth()==8)
	{
		sphere=CreateSurface( "sphere_08.bmp", TRUE );
		if (sphere==0)
			Fatal("Can't find sphere_08.bmp");
	}
	else
	{
		sphere=CreateSurface( "sphere_24.bmp" );
		if (sphere==0)
			Fatal("Can't find sphere_24.bmp");
	}

	DDCOLORKEY  colorkey;
	colorkey.dwColorSpaceLowValue = 0;
	colorkey.dwColorSpaceHighValue = 0;
	sphere->SetColorKey( DDCKEY_SRCBLT, &colorkey );

	ClearSurface( primsurf, 0 );

	return TRUE;
}

void SmearWin::DrawScene()
{
	static char key[256];
	keyboard->GetDeviceState( sizeof(key), &key );

	if ( key[DIK_ESCAPE] & 0x80 )
		PostMessage( WM_CLOSE );

	BOOL done=FALSE;
	while (!done)
	{
		DIDEVICEOBJECTDATA data;
		DWORD elements=1;
		HRESULT r=mouse->GetDeviceData( sizeof(data), &data, &elements, 0 );
		if (r==DI_OK && elements==1)
		{
			switch(data.dwOfs)
			{
	        case DIMOFS_X:
				x+=data.dwData;
				break;
		    case DIMOFS_Y: 
				y+=data.dwData;
				break;
			}
		}
		else if (elements==0)
			done=TRUE;
	}

	BltSurface( primsurf, sphere, x, y, TRUE );
}

void SmearWin::RestoreSurfaces()
{
	if (GetCurDisplayDepth()==8)
	{
		sphere->Restore();
		LoadSurface( sphere, "sphere_08.bmp" );
	}
	else
	{
		sphere->Restore();
		LoadSurface( sphere, "sphere_24.bmp" );
	}
}

int SmearWin::SelectDriver()
{
	int numdrivers=GetNumDrivers();
	if (numdrivers==1)
		return 0;

	CArray<CString, CString> drivers;
	for (int i=0;i<numdrivers;i++)
	{
		LPSTR desc, name;
		GetDriverInfo( i, 0, &desc, &name );
		drivers.Add(desc);
	}

	DriverDialog dialog;
	dialog.SetContents( &drivers );
	if (dialog.DoModal()!=IDOK)
		return -1;

	return dialog.GetSelection();
}

int SmearWin::SelectInitialDisplayMode()
{
	DWORD curdepth=GetDisplayDepth();
	int i, nummodes=GetNumDisplayModes();
	DWORD w,h,d;

	//----- look for 640x480 with current pixel depth
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==640 && h==480 && d==curdepth)
			return i;
	}

	//----- barring that, try anything with the current depth
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (d==curdepth)
			return i;
	}

	//----- if that fails, try finding any 640x480 mode
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==640 && h==480)
			return i;
	}

	return 0;
}

int SmearWin::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	HRESULT r=DirectInputCreate( AfxGetInstanceHandle(), 
			DIRECTINPUT_VERSION, &dinput, 0 ); 
	if (r!=DI_OK)
	{
		AfxMessageBox("DirectInputCreate() failed");
		return -1;
	}

	if (InitMouse()==FALSE)
		return -1;

	if (InitKeyboard()==FALSE)
		return -1;

	if (DirectDrawWin::OnCreate(lpCreateStruct) == -1)
		return -1;

	return 0;
}

void SmearWin::OnDestroy() 
{
	DirectDrawWin::OnDestroy();
	
	if (dinput)
		dinput->Release(), dinput=0;

	if (keyboard)
	{
		keyboard->Unacquire();
		keyboard->Release(), keyboard=0;
	}

	if (mouse)
	{
		mouse->Unacquire();
		mouse->Release(), mouse=0;
	}
}

void SmearWin::OnActivate(UINT nState, CWnd* pWndOther, BOOL bMinimized) 
{
	DirectDrawWin::OnActivate(nState, pWndOther, bMinimized);
	
	if (nState!=WA_INACTIVE)
	{
		if (keyboard)
		{
			TRACE("keyboard->Acquire()\n");
			keyboard->Acquire();
		}
		if (mouse)
		{
			TRACE("mouse->Acquire()\n");
			mouse->Acquire();
		}
	}
}

BOOL SmearWin::InitKeyboard()
{
	HRESULT r;
	r = dinput->CreateDevice( GUID_SysKeyboard, &keyboard, 0 ); 
	if (r!=DI_OK)
	{
		TRACE("CreateDevice(keyboard) failed");	
		return FALSE;
	}
	
	r = keyboard->SetDataFormat( &c_dfDIKeyboard );
	if (r!=DI_OK)
	{
		TRACE("keyboard->SetDataFormat() failed\n");
		return FALSE;
	}

	r=keyboard->SetCooperativeLevel( GetSafeHwnd(), 
			DISCL_FOREGROUND | DISCL_NONEXCLUSIVE); 
	if (r!=DI_OK)
	{
		TRACE("keyboard->SetCooperativeLevel() failed\n");
		return FALSE;
	}

	return TRUE;
}

BOOL SmearWin::InitMouse()
{
	HRESULT r;
	r = dinput->CreateDevice( GUID_SysMouse, &mouse, 0 ); 
	if (r!=DI_OK)
	{
		TRACE("CreateDevice(mouse) failed\n");	
		return FALSE;
	}

	r = mouse->SetDataFormat( &c_dfDIMouse );
	if (r!=DI_OK)
	{
		TRACE("mouse->SetDataFormat() failed\n");
		return FALSE;
	}

	r = mouse->SetCooperativeLevel( GetSafeHwnd(),
			DISCL_NONEXCLUSIVE | DISCL_FOREGROUND );
	if (r!=DI_OK)
	{
		TRACE("mouse->SetCooperativeLevel() failed\n");
		return FALSE;
	}

	DIPROPDWORD property;
	property.diph.dwSize=sizeof(DIPROPDWORD);
	property.diph.dwHeaderSize=sizeof(DIPROPHEADER);
	property.diph.dwObj=0;
	property.diph.dwHow=DIPH_DEVICE;
	property.dwData=64;

	r = mouse->SetProperty( DIPROP_BUFFERSIZE, &property.diph );
	if (r!=DI_OK)
	{
		TRACE("mouse->SetProperty() failed (buffersize)\n");
		return FALSE;
	}

	return TRUE;
}

BOOL SmearWin::CreateFlippingSurfaces()
{
	if (primsurf)
		primsurf->Release(), primsurf=0;

	DDSURFACEDESC desc;
	ZeroMemory( &desc, sizeof(desc) );
	desc.dwSize = sizeof( desc );
	desc.dwFlags =  DDSD_CAPS;
	desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	HRESULT r=ddraw2->CreateSurface( &desc, &primsurf, 0 );
	if (r!=DD_OK)
		return FALSE;

	backsurf=0;

	return TRUE;
}