#include "Headers.h"
#include "resource.h"
#include "DriverDialog.h"
#include "IntroDialog.h"
#include "SuperSwitchWin.h"

#pragma comment (lib,"ddraw.lib")
#pragma comment (lib,"dxguid.lib")
#pragma comment (lib,"winmm.lib")


const int menucols = 4;
const int reservedspace=30;
const int menuwidth=320;
const int menuheight=200;
const int colwidth=80;
const int rowheight=15;
const int maxmodes=40;
const char modeheader[]="SuperSwitch Demo";
const char rateheader[]="Refresh rates";

const COLORREF fpstextcolor=RGB(144, 120, 120);
const COLORREF fpstextshadow=RGB(50, 50, 50);
const COLORREF fpsbrighttextcolor=RGB(255, 255, 255);

const COLORREF modetextcolor=RGB(144, 120, 120);
const COLORREF modetextshadow=RGB(50, 50, 50);
const COLORREF modehighlightcolor=RGB(255, 255, 255);

const COLORREF ratetextcolor=RGB(0, 200, 132);
const COLORREF ratetextshadow=RGB(0, 75, 0);
const COLORREF ratehighlightcolor=RGB(255, 255, 255);


BEGIN_MESSAGE_MAP(SuperSwitchWin, DirectDrawWin)
	//{{AFX_MSG_MAP(SuperSwitchWin)
	ON_WM_KEYDOWN()
	ON_WM_CREATE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


BOOL SuperSwitchWin::CreateCustomSurfaces()
{
	int displaydepth=GetCurDisplayDepth();

	CString filename;

	if (displaydepth==8)
		filename="tri08.bmp";
	else
		filename="tri24.bmp";

	bmpsurf=CreateSurface( filename, TRUE );
	if (bmpsurf==0)
	{
		CString msg="Can't open file: ";
		msg+=filename;
		Fatal( msg );
	}

	selectmode=GetCurDisplayMode();
	CreateModeMenuSurface();
	UpdateModeMenuSurface();

	selectrate=0;

	CreateFPSSurface();

	return TRUE;
}

BOOL SuperSwitchWin::CreateModeMenuSurface()
{
	if (modemenusurf)
		modemenusurf->Release(), modemenusurf=0;

	modemenusurf=CreateSurface( menuwidth, menuheight );
	if (modemenusurf==0)
		Fatal("SuperSwitchWin::CreateModeMenuSurface() failed\n");

	DDCOLORKEY  ddck;
	ddck.dwColorSpaceLowValue = 0;
	ddck.dwColorSpaceHighValue = 0;

	modemenusurf->SetColorKey( DDCKEY_SRCBLT, &ddck );

	return TRUE;
}

BOOL SuperSwitchWin::CreateRateMenuSurface()
{
	if (ratemenusurf)
		ratemenusurf->Release(), ratemenusurf=0;

	int rates=refresh_rates[selectmode].GetSize();

	ratemenusurf=CreateSurface( 80, rates*12+22 );

	return TRUE;
}

void SuperSwitchWin::DrawScene()
{
	ClearSurface( backsurf, 0 );

	BltSurface( backsurf, bmpsurf, x, y );

	x+=xinc;
	y+=yinc;

	const CRect& displayrect=GetDisplayRect();
	if (x<-160 || x>displayrect.right-160)
	{
		xinc=-xinc;
		x+=xinc;
	}
	if (y<-100 || y>displayrect.bottom-100)
	{
		yinc=-yinc;
		y+=yinc;
	}

	backsurf->BltFast( 0, 0, modemenusurf, 0, 
			DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT );

	if (ratemenu_up)
	{
		DWORD w,h;
		GetSurfaceDimensions( ratemenusurf, w, h );
		backsurf->BltFast( (320-w)/2, (200-h)/2, ratemenusurf, 0, 
				DDBLTFAST_WAIT );
	}

	UpdateFPSSurface();
	if (displayfps)
	{
		int x=displayrect.right-fpsrect.right;
		int y=displayrect.bottom-fpsrect.bottom;
		backsurf->BltFast( x, y, fpssurf, &fpsrect,
				DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT );
	}

	primsurf->Flip( 0, DDFLIP_WAIT );
}

int SuperSwitchWin::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	IntroDialog introdialog;
	if (introdialog.DoModal()!=IDOK)
		return -1;

	include_refresh=introdialog.include_refresh;

	if (DirectDrawWin::OnCreate(lpCreateStruct)==-1)
		return -1;

	if (include_refresh)
		ddraw2->EnumDisplayModes( DDEDM_REFRESHRATES, 0, this, 
				StoreModeInfo );

	return 0;
}

void SuperSwitchWin::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	int newindex;
	int nmodes=GetNumDisplayModes();
	if (nmodes>maxmodes)
		nmodes=maxmodes;
	int rows=nmodes/menucols;
	if (nmodes%menucols)
		rows++;

	switch (nChar)
	{
	case VK_ESCAPE:
		if (!include_refresh || !ratemenu_up)
		{
			PostMessage( WM_CLOSE );
			break;
		}
		if (ratemenu_up)
		{
			ratemenu_up=FALSE;
			if (ratemenusurf)
				ratemenusurf->Release(), ratemenusurf=0;
		}
		break;
	case VK_UP:
		if ( include_refresh && ratemenu_up)
		{
			if (selectrate>0)
			{
				selectrate--;
				UpdateRateMenuSurface();
			}
		}
		else
		{
			newindex=selectmode-1;
			if (newindex>=0)
			{
				selectmode=newindex;
				UpdateModeMenuSurface();
			}
		}
		break;
	case VK_DOWN:
		if (include_refresh && ratemenu_up)
		{
			if (selectrate<numrates-1)
			{
				selectrate++;
				UpdateRateMenuSurface();
			}
		}
		else
		{
			newindex=selectmode+1;
			if (newindex<nmodes)
			{
				selectmode=newindex;
				UpdateModeMenuSurface();
			}
		}
		break;
	case VK_LEFT:
		if (include_refresh && ratemenu_up)
			break;
		newindex=selectmode-rows;
		if (newindex>=0)
		{
			selectmode=newindex;
			UpdateModeMenuSurface();
		}
		break;
	case VK_RIGHT:
		if (include_refresh && ratemenu_up)
			break;
		newindex=selectmode+rows;
		if (newindex<nmodes)
		{
			selectmode=newindex;
			UpdateModeMenuSurface();
		}
		break;
	case VK_RETURN:
		if (include_refresh)
		{
			if (ratemenu_up)
			{
				int rate=refresh_rates[selectmode][selectrate];
				ActivateDisplayMode( selectmode, rate );
				x=y=0;
				ratemenu_up=FALSE;
			}
			else
			{
				CreateRateMenuSurface();
				UpdateRateMenuSurface();
				ratemenu_up=TRUE;
			}
		}
		else
		{
			if (selectmode!=GetCurDisplayMode())
			{
				ActivateDisplayMode( selectmode );
				x=y=0;
			}
		}
		break;
	case 'S':
		SaveSurface( primsurf, "SuperSwitch.bmp" );
		break;
	default:
		DirectDrawWin::OnKeyDown(nChar, nRepCnt, nFlags);
	}
	
	
}

void SuperSwitchWin::RestoreSurfaces()
{
	int displaydepth=GetCurDisplayDepth();

	CString filename;
	if (displaydepth==8)
		filename="tri08.bmp";
	else
		filename="tri24.bmp";

	if (bmpsurf->IsLost())
	{
		bmpsurf->Restore();
		LoadSurface( bmpsurf, filename );
	}

	if (modemenusurf->IsLost())
	{
		modemenusurf->Restore();
		UpdateModeMenuSurface();
	}

	if (ratemenu_up && ratemenusurf->IsLost())
	{
		ratemenusurf->Restore();
		UpdateRateMenuSurface();
	}

	if (fpssurf->IsLost())
	{
		fpssurf->Restore();
		ClearSurface( fpssurf, 0 );
	}
	displayfps=FALSE;
	framecount=0;
}

int SuperSwitchWin::SelectDriver()
{
	int numdrivers=GetNumDrivers();
	if (numdrivers==1)
		return 0;

	CArray<CString, CString> drivers;
	for (int i=0;i<numdrivers;i++)
	{
		LPSTR desc, name;
		GetDriverInfo( i, 0, &desc, &name );
		drivers.Add(desc);
	}

	DriverDialog dialog;
	dialog.SetContents( &drivers );
	if (dialog.DoModal()!=IDOK)
		return -1;

	return dialog.GetSelection();
}

int SuperSwitchWin::SelectInitialDisplayMode()
{
	DWORD curdepth=GetDisplayDepth();
	int i, nummodes=GetNumDisplayModes();
	DWORD w,h,d;

	//----- look for 640x480 with current pixel depth
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==640 && h==480 && d==curdepth)
			return i;
	}

	//----- barring that, try anything with the current depth
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (d==curdepth)
			return i;
	}

	//----- if that fails, try finding any 640x480 mode
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==640 && h==480)
			return i;
	}

	return 0;
}

HRESULT WINAPI SuperSwitchWin::StoreModeInfo( LPDDSURFACEDESC desc, LPVOID p)
{
	DWORD w=desc->dwWidth;
	DWORD h=desc->dwHeight;
	DWORD d=desc->ddpfPixelFormat.dwRGBBitCount;
	DWORD r=desc->dwRefreshRate;
	
	SuperSwitchWin* win=(SuperSwitchWin*)p;
	int index=win->GetDisplayModeIndex( w, h, d );
	win->refresh_rates[index].Add(r);

	return DDENUMRET_OK;
}

SuperSwitchWin::SuperSwitchWin()
{
	modemenusurf=0;

	ratemenusurf=0;
	ratemenu_up=FALSE;

	bmpsurf=0;
	x=y=0;
	xinc=8;
	yinc=1;

	largefont = CreateFont( 28, 0, 0, 0,
		FW_NORMAL, FALSE, FALSE, FALSE,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		VARIABLE_PITCH,
		"Arial" );

	smallfont = CreateFont( 14, 0, 0, 0,
		FW_NORMAL, FALSE, FALSE, FALSE,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		VARIABLE_PITCH,
		"Arial" );
}

BOOL SuperSwitchWin::UpdateModeMenuSurface()
{
	ClearSurface( modemenusurf, 0 );

	HDC hdc;
	modemenusurf->GetDC( &hdc );
	SelectObject( hdc, largefont );
	
	SetBkMode( hdc, TRANSPARENT );

	SetTextColor( hdc, modetextshadow );
	ExtTextOut(hdc, 1, 1, 0, 0, modeheader, strlen(modeheader), 0 );
	SetTextColor( hdc, modetextcolor );
	ExtTextOut(hdc, 0, 0, 0, 0, modeheader, strlen(modeheader), 0 );

	SelectObject( hdc, smallfont );

	int nmodes=GetNumDisplayModes();
	if (nmodes>maxmodes)
		nmodes=maxmodes;

	int rows=nmodes/menucols;
	if (nmodes%menucols)
		rows++;

	char buf[30];
	RECT rect;
	for (int i=0; i<nmodes; i++)
	{
		rect.left=(i/rows)*colwidth;
		rect.top=(i%rows)*rowheight+reservedspace;
		rect.right=rect.left+colwidth;
		rect.bottom=rect.top+rowheight;
	
		DWORD w,h,d;
		GetDisplayModeDimensions( i, w, h, d );

		int len=sprintf( buf, "%dx%dx%d", w, h, d );

		SetTextColor( hdc, modetextshadow );
		ExtTextOut(hdc, rect.left+1, rect.top+1, 0, &rect, buf, len, 0 );

		if (i==selectmode)
			SetTextColor( hdc, modehighlightcolor );
		else
			SetTextColor( hdc, modetextcolor );
		ExtTextOut(hdc, rect.left, rect.top, 0, &rect, buf, len, 0 );
	}

	rect.left=0;
	rect.top=179;
	rect.right=319;
	rect.bottom=199;

	int len=sprintf( buf, "[Arrows]  [Enter]  [Escape]" );
	SetTextColor( hdc, modetextshadow );
	ExtTextOut(hdc, 1, 180, 0, &rect, buf, len, 0 );
	SetTextColor( hdc, modetextcolor );
	ExtTextOut(hdc, 0, 179, 0, &rect, buf, len, 0 );
	
	modemenusurf->ReleaseDC( hdc );

	return TRUE;
}

BOOL SuperSwitchWin::UpdateRateMenuSurface()
{
	RECT rect;
	GetSurfaceRect( ratemenusurf, rect );
	rect.left++;
	rect.top++;
	rect.right--;
	rect.bottom--;

	if (!ClearSurface( ratemenusurf, 0, 200, 132 ))
		TRACE("first Clear failed\n");
	if (!ClearSurface( ratemenusurf, 0,128,100, &rect ))
		TRACE("second Clear failed\n");

	HDC hdc;
	ratemenusurf->GetDC( &hdc );
	SelectObject( hdc, smallfont );
	SetBkMode( hdc, TRANSPARENT );

	SetTextColor( hdc, ratetextshadow );
	ExtTextOut(hdc, 6, 4, 0, 0, rateheader, strlen(rateheader), 0 );
	SetTextColor( hdc, ratetextcolor );
	ExtTextOut(hdc, 5, 3, 0, 0, rateheader, strlen(rateheader), 0 );

	CArray<DWORD,DWORD>& ratelist=refresh_rates[selectmode];
	numrates=ratelist.GetSize();
	for (int i=0; i<numrates; i++)
	{
		char buf[10];
		int len=sprintf( buf, "%d hz", ratelist[i] );

		SetTextColor( hdc, ratetextshadow );
		ExtTextOut(hdc, 11, i*12+18, 0, 0, buf, len, 0 );

		if (i==selectrate)
			SetTextColor( hdc, ratehighlightcolor );
		else
			SetTextColor( hdc, ratetextcolor );
		ExtTextOut(hdc, 10, i*12+17, 0, 0, buf, len, 0 );
	}

	ratemenusurf->ReleaseDC( hdc );

	return TRUE;
}

BOOL SuperSwitchWin::UpdateFPSSurface()
{
	static const long interval=100;
	framecount++;

	if (framecount==interval)
	{
		static DWORD timenow;
		static DWORD timethen;
		timethen=timenow;
		timenow=timeGetTime();

		double seconds=double(timenow-timethen)/(double)1000;
		int fps=(int)((double)framecount/seconds);

		static char buf[10];
		int len=sprintf( buf, "%d FPS", fps);

		ClearSurface( fpssurf, 0 );
		
		HDC hdc;
		fpssurf->GetDC( &hdc );

		SelectObject( hdc, smallfont );
		SetBkMode( hdc, TRANSPARENT );
		SetBkColor( hdc, RGB(0,0,0) );

		SetTextColor( hdc, fpstextshadow );
		TextOut(hdc, 1, 1, buf, len );

		SetTextColor( hdc, fpsbrighttextcolor );
		TextOut(hdc, 0, 0, buf, len );
		
		fpssurf->ReleaseDC( hdc );

		displayfps=TRUE;
		framecount=0;
	}

	return TRUE;
}

BOOL SuperSwitchWin::CreateFPSSurface()
{
	static const char dummystr[]="000 FPS";

	HDC hdc = ::GetDC( 0 );
	SelectObject( hdc, smallfont );
	SIZE size;
	GetTextExtentPoint( hdc, dummystr, strlen(dummystr), &size );
	::ReleaseDC( 0, hdc );

	fpsrect.left=0;
	fpsrect.top=0;
	fpsrect.right=size.cx+1;
	fpsrect.bottom=size.cy+1;

	fpssurf=CreateSurface( fpsrect.right, fpsrect.bottom );
	ClearSurface( fpssurf, 0 );

	DDCOLORKEY  ddck;
	ddck.dwColorSpaceLowValue = 0;
	ddck.dwColorSpaceHighValue = 0;
	fpssurf->SetColorKey( DDCKEY_SRCBLT, &ddck );

	framecount=0;

	return TRUE;
}
