#include "Headers.h"
#include "resource.h"
#include "DriverDialog.h"
#include "SwitchWin.h"

#pragma comment (lib,"ddraw.lib")
#pragma comment (lib,"dxguid.lib")
#pragma comment (lib,"winmm.lib")


const int menucols = 4;
const int reservedspace=30;
const int menuwidth=320;
const int menuheight=200;
const int colwidth=80;
const int rowheight=15;
const int maxmodes=40;
const char headertext[]="DirectDraw Switch Demo";
const COLORREF textcolor=RGB(144, 120, 120);
const COLORREF textshadow=RGB(50, 50, 50);
const COLORREF brighttextcolor=RGB(255, 255, 255);


BEGIN_MESSAGE_MAP(SwitchWin, DirectDrawWin)
	//{{AFX_MSG_MAP(SwitchWin)
	ON_WM_KEYDOWN()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


BOOL SwitchWin::CreateCustomSurfaces()
{
	//take this away
	int displaydepth=GetCurDisplayDepth();
	CString filename;
	if (displaydepth==8)
		filename="tri08.bmp";
	else
		filename="tri24.bmp";

	bmpsurf=CreateSurface( filename, TRUE );
	if (bmpsurf==0)
	{
		CString msg="Can't open file: ";
		msg+=filename;
		Fatal( msg );
	}

	selectmode=GetCurDisplayMode();
	CreateMenuSurface();
	UpdateMenuSurface();

	CreateFPSSurface();

	return TRUE;
}

BOOL SwitchWin::CreateFPSSurface()
{
	static const char dummystr[]="000 FPS";

	HDC hdc = ::GetDC( 0 );
	SelectObject( hdc, smallfont );
	SIZE size;
	GetTextExtentPoint( hdc, dummystr, strlen(dummystr), &size );
	::ReleaseDC( 0, hdc );

	fpsrect.left=0;
	fpsrect.top=0;
	fpsrect.right=size.cx+1;
	fpsrect.bottom=size.cy+1;

	TRACE("fpssurf: %dx%d\n", fpsrect.right, fpsrect.bottom);
	fpssurf=CreateSurface( fpsrect.right, fpsrect.bottom );

	DDCOLORKEY  ddck;
	ddck.dwColorSpaceLowValue = 0;
	ddck.dwColorSpaceHighValue = 0;
	fpssurf->SetColorKey( DDCKEY_SRCBLT, &ddck );

	framecount=0;
	displayfps=FALSE;

	return TRUE;
}

BOOL SwitchWin::CreateMenuSurface()
{
	if (menusurf)
		menusurf->Release(), menusurf=0;

	menusurf=CreateSurface( menuwidth, menuheight );
	if (menusurf==0)
		Fatal("SwitchWin::CreateMenuSurface() failed\n");

	DDCOLORKEY  ddck;
	ddck.dwColorSpaceLowValue = 0;
	ddck.dwColorSpaceHighValue = 0;
	menusurf->SetColorKey( DDCKEY_SRCBLT, &ddck );

	return TRUE;
}

void SwitchWin::DrawScene()
{
	ClearSurface( backsurf, 0 );

	BltSurface( backsurf, bmpsurf, x, y );

	x+=xinc;
	y+=yinc;

	const CRect& displayrect=GetDisplayRect();
	if (x<-160 || x>displayrect.right-160)
	{
		xinc=-xinc;
		x+=xinc;
	}
	if (y<-100 || y>displayrect.bottom-100)
	{
		yinc=-yinc;
		y+=yinc;
	}

	backsurf->BltFast( 0, 0, menusurf, 0, 
			DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT );

	UpdateFPSSurface();
	if (displayfps)
	{
		int x=displayrect.right-fpsrect.right-1;
		int y=displayrect.bottom-fpsrect.bottom-1;
		backsurf->BltFast( x, y, fpssurf, 0,
				DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT );
	}

	primsurf->Flip( 0, DDFLIP_WAIT );
}

void SwitchWin::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	int newindex;
	int nmodes=GetNumDisplayModes();
	if (nmodes>maxmodes)
		nmodes=maxmodes;
	int rows=nmodes/menucols;
	if (nmodes%menucols)
		rows++;

	switch (nChar)
	{
	case VK_ESCAPE:
		PostMessage( WM_CLOSE );
		break;
	case VK_UP:
		newindex=selectmode-1;
		if (newindex>=0)
		{
			selectmode=newindex;
			UpdateMenuSurface();
		}
		break;
	case VK_DOWN:
		newindex=selectmode+1;
		if (newindex<nmodes)
		{
			selectmode=newindex;
			UpdateMenuSurface();
		}
		break;
	case VK_LEFT:
		newindex=selectmode-rows;
		if (newindex>=0)
		{
			selectmode=newindex;
			UpdateMenuSurface();
		}
		break;
	case VK_RIGHT:

		newindex=selectmode+rows;
		if (newindex<nmodes)
		{
			selectmode=newindex;
			UpdateMenuSurface();
		}
		break;
	case VK_RETURN:
		if (selectmode != GetCurDisplayMode())
		{
			ActivateDisplayMode( selectmode );
			x=y=0;
		}
		break;
	case 'S':
		SaveSurface( primsurf, "switch.bmp" );
		break;
	case 'M':
		SaveSurface( menusurf, "menusurf.bmp" );
		break;
	case 'F':
		SaveSurface( fpssurf, "fpssurf.bmp" );
		break;
	case 'T':
		SaveSurface( bmpsurf, "trisurf.bmp" );
		break;
	default:
		DirectDrawWin::OnKeyDown(nChar, nRepCnt, nFlags);
	}
}

void SwitchWin::RestoreSurfaces()
{
	int displaydepth=GetCurDisplayDepth();

	CString filename;
	if (displaydepth==8)
		filename="tri08.bmp";
	else
		filename="tri24.bmp";

	if (bmpsurf->IsLost())
	{
		bmpsurf->Restore();
		LoadSurface( bmpsurf, filename );
	}

	if (menusurf->IsLost())
	{
		menusurf->Restore();
		UpdateMenuSurface();
	}

	if (fpssurf->IsLost())
	{
		fpssurf->Restore();
		ClearSurface( fpssurf, 0 );
	}
	displayfps=FALSE;
	framecount=0;
}

int SwitchWin::SelectDriver()
{
	int numdrivers=GetNumDrivers();
	if (numdrivers==1)
		return 0;

	CArray<CString, CString> drivers;
	for (int i=0;i<numdrivers;i++)
	{
		LPSTR desc, name;
		GetDriverInfo( i, 0, &desc, &name );
		drivers.Add(desc);
	}

	DriverDialog dialog;
	dialog.SetContents( &drivers );
	if (dialog.DoModal()!=IDOK)
		return -1;

	return dialog.GetSelection();
}

int SwitchWin::SelectInitialDisplayMode()
{
	DWORD curdepth=GetDisplayDepth();
	int i, nummodes=GetNumDisplayModes();
	DWORD w,h,d;

	//----- look for 640x480 with current pixel depth
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==640 && h==480 && d==curdepth)
			return i;
	}

	//----- barring that, try anything with the current depth
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (d==curdepth)
			return i;
	}

	//----- if that fails, try finding any 640x480 mode
	for (i=0;i<nummodes;i++)
	{
		GetDisplayModeDimensions( i, w, h, d );
		if (w==640 && h==480)
			return i;
	}

	return 0;
}

SwitchWin::SwitchWin()
{
	bmpsurf=0;
	x=y=0;
	xinc=8;
	yinc=1;

	menusurf=0;
	fpssurf=0;

	largefont = CreateFont( 28, 0, 0, 0,
		FW_NORMAL, FALSE, FALSE, FALSE,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		VARIABLE_PITCH,
		"Arial" );

	smallfont = CreateFont( 14, 0, 0, 0,
		FW_NORMAL, FALSE, FALSE, FALSE,
		ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		VARIABLE_PITCH,
		"Arial" );
}

BOOL SwitchWin::UpdateFPSSurface()
{
	static const long interval=100;
	framecount++;

	if (framecount==interval)
	{
		static DWORD timenow;
		static DWORD timethen;
		timethen=timenow;
		timenow=timeGetTime();

		double seconds=double(timenow-timethen)/(double)1000;
		int fps=(int)((double)framecount/seconds);

		static char buf[10];
		int len=sprintf( buf, "%d FPS", fps);

		ClearSurface( fpssurf, 0 );
		
		HDC hdc;
		fpssurf->GetDC( &hdc );

		SelectObject( hdc, smallfont );
		SetBkMode( hdc, TRANSPARENT );
		SetBkColor( hdc, RGB(0,0,0) );

		SetTextColor( hdc, textshadow );
		TextOut(hdc, 1, 1, buf, len );

		SetTextColor( hdc, brighttextcolor );
		TextOut(hdc, 0, 0, buf, len );
		
		fpssurf->ReleaseDC( hdc );

		displayfps=TRUE;
		framecount=0;
	}

	return TRUE;
}

BOOL SwitchWin::UpdateMenuSurface()
{
	char buf[40];
	int len;
	int hdrlen=strlen( headertext );
		
	ClearSurface( menusurf, 0 );

	HDC hdc;
	menusurf->GetDC( &hdc );
	SelectObject( hdc, largefont );

	SetBkMode( hdc, TRANSPARENT );

	SetTextColor( hdc, textshadow );
	TextOut( hdc, 1, 1, headertext, hdrlen );

	SetTextColor( hdc, textcolor );
	TextOut( hdc, 0, 0, headertext, hdrlen );

	SelectObject( hdc, smallfont );

	int nmodes=GetNumDisplayModes();
	if (nmodes>maxmodes)
		nmodes=maxmodes;

	int rows=nmodes/menucols;
	if (nmodes%menucols)
		rows++;

	for (int i=0; i<nmodes; i++)
	{
		int x=(i/rows)*colwidth+2;
		int y=(i%rows)*rowheight+reservedspace;
	
		DWORD w,h,d;
		GetDisplayModeDimensions( i, w, h, d );

		len=sprintf( buf, "%dx%dx%d", w, h, d );

		SetTextColor( hdc, textshadow );
		TextOut( hdc, x+1, y+1, buf, len );

		if (i==selectmode)
			SetTextColor( hdc, brighttextcolor );
		else
			SetTextColor( hdc, textcolor );
		TextOut( hdc, x, y, buf, len );
	}

	len=sprintf( buf, "[Arrows]  [Enter]  [Escape]" );
	SetTextColor( hdc, textshadow );
	TextOut( hdc, 3, 186, buf, len );
	SetTextColor( hdc, textcolor );
	TextOut( hdc, 2, 185, buf, len );
	
	menusurf->ReleaseDC( hdc );

	return TRUE;
}

